"""Authentication handling in the form of token-providers"""

import base64
import json
import logging
import re
import sys
import time
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, List, Optional, Union

import jwt
import msal
import msal_extensions
import requests
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicNumbers
from requests.auth import AuthBase

from ..util.cheapdantic import BaseModel, PrivateAttr, SecretStr
from ..util.check_version import get_sdk_version


class OdpAuthError(Exception):
    """Exception raised for authentication errors in ODP SDK"""


LOG = logging.getLogger(__name__)


class JwtTokenProvider(BaseModel, AuthBase, ABC):
    """Token provider for JWT tokens"""

    user_agent: str = "odp-sdk/" + get_sdk_version()

    audience: Optional[str] = None
    """IDP token audience

    Must be set if the token is to be validated
    """

    validate_token: bool = False
    """Whether to validate the token or not"""

    token_exp_lee_way: int = 300
    """Number of seconds before token expiry we should refresh the token"""

    user_id_claim: str = "sub"
    """The claim to use as the user ID"""

    _access_token: Optional[str] = PrivateAttr(None)
    _refresh_token: Optional[str] = PrivateAttr(None)
    _claims: Dict[str, Union[int, str]] = PrivateAttr(None)
    _jwks: Dict[str, str] = PrivateAttr(None)
    _expiry: int = PrivateAttr(0)
    _user_id: Optional[str] = PrivateAttr(None)

    def __call__(self, r: requests.PreparedRequest) -> requests.PreparedRequest:
        r.headers["Authorization"] = self.get_token()
        return r

    @abstractmethod
    def authenticate(self) -> Dict[str, str]:
        """Authenticate against the IDP and return the token

        Returns:
            The token response from the IDP

        Raises:
            OdpAuthError: If the token cannot be retrieved
        """

    @abstractmethod
    def get_jwks_uri(self) -> str:
        """Get the jwks uri for the identity provider

        Returns:
            The jwks uri
        """

    def get_token(self) -> str:
        """Returns the token to be used for authentication

        Returns:
            The token to be used for authentication

        Raises:
            OdpAuthError: If the token cannot be retrieved
        """

        if self._access_token and time.time() < self._expiry - self.token_exp_lee_way:
            access_token = self._access_token
        else:
            auth_response = self.authenticate()
            access_token = self._parse_token(auth_response)
            self._user_id = self._claims[self.user_id_claim]

        return "Bearer {}".format(access_token)

    def get_user_id(self) -> str:
        if not self._user_id:
            self.get_token()  # This will set the user_id
            if not self._user_id:
                raise OdpAuthError("Unable to obtain user ID")

        return self._user_id

    def _parse_and_validate(self, access_token: str) -> str:
        token_components = [self._base64_decode(x) for x in access_token.split(".")]
        headers = json.loads(token_components[0])

        if not headers:
            raise OdpAuthError("No headers in token")

        kid = headers.pop("kid")
        jwk = self._get_jwk(kid)

        pem = (
            RSAPublicNumbers(n=self._decode_jwk_value(jwk["n"]), e=self._decode_jwk_value(jwk["e"]))
            .public_key(default_backend())
            .public_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PublicFormat.SubjectPublicKeyInfo,
            )
        )

        self._claims = jwt.decode(
            access_token,
            pem,
            verify=True,
            options=headers,
            audience=self.audience,
            algorithms=["RS256"],
        )

        self._access_token = access_token
        self._expiry = self._claims["exp"]

        return access_token

    def _parse_novalidate(self, access_token: str) -> str:
        self._claims = jwt.decode(access_token, options={"verify_signature": False})
        self._access_token = access_token
        self._expiry = self._claims["exp"]

        return access_token

    def _parse_token(self, token_response: Dict[str, str]) -> str:
        """Parse the token from the token response

        Args:
            The token response from the IDP

        Returns:
            str: The token
        """
        try:
            access_token = token_response["access_token"]
        except KeyError as e:
            raise OdpAuthError("No access token in token response") from e

        if not self.validate_token:
            return self._parse_novalidate(access_token)
        return self._parse_and_validate(access_token)

    def _get_jwk(self, kid: str) -> Dict:
        jwks = self._get_jwks()
        for jwk in jwks["keys"]:
            if jwk.get("kid") == kid:
                return jwk

        raise OdpAuthError("Invalid token: Invalid KID")

    def _get_jwks(self) -> Dict:
        if self._jwks:
            return self._jwks

        jwks_uri = self.get_jwks_uri()

        res = requests.get(jwks_uri)
        res.raise_for_status()

        self._jwks = res.json()
        return self._jwks

    @staticmethod
    def _base64_decode(data: str, altchars: bytes = b"+/") -> bytes:
        d = re.sub(rb"[^a-zA-Z0-9%s]+" % altchars, b"", data.encode("utf-8"))

        if len(d) % 4 == 1:
            d = d[:-1]
        else:
            # Add padding
            d += (-len(d) % 4) * b"="
        decoded = base64.b64decode(d, altchars=altchars)
        return decoded

    @staticmethod
    def _decode_jwk_value(val: str) -> int:
        decoded = base64.urlsafe_b64decode(val.encode("utf-8") + b"==")
        return int.from_bytes(decoded, "big")


class InteractiveTokenProvider(JwtTokenProvider):
    """Token provider for interactive token input

    This token provider is intended to be used for interactive sessions, where the user is prompted for
    login credentials by a web browser. The token is then retrieved from the browser and used for authentication.
    """

    client_id: SecretStr
    """IDP client ID"""

    authority: str = (
        "https://oceandataplatform.b2clogin.com/oceandataplatform.onmicrosoft.com/b2c_1a_signup_signin_custom"
    )
    """IDP token Authority"""

    scope: List[str] = ["https://oceandataplatform.onmicrosoft.com/odcat/API_ACCESS"]
    """IDP token scope"""

    jwks_uri: str = "https://oceandataplatform.b2clogin.com/oceandataplatform.onmicrosoft.com/b2c_1a_signup_signin_custom/discovery/v2.0/keys"  # noqa: E501
    """JWKS endpoint."""

    token_persistence_file: Path = Path(".token_cache.bin")
    """File to save the token cache to"""

    token_persistence_plaintext_fallback: bool = True
    """Whether to fallback to plaintext if encryption is unavailable"""

    _app: msal.PublicClientApplication = PrivateAttr(None)

    def __init__(self, **data):
        super().__init__(**data)

        persistence = self._build_persistence(self.token_persistence_file, self.token_persistence_plaintext_fallback)
        cache = msal_extensions.PersistedTokenCache(persistence)

        self._app = msal.PublicClientApplication(
            client_id=self.client_id.get_secret_value(),
            authority=self.authority,
            token_cache=cache,
        )
        self.user_agent = self.user_agent + " (Interactive)"

    def get_jwks_uri(self) -> str:
        return self.jwks_uri

    def authenticate(self) -> Dict[str, str]:
        accounts = self._app.get_accounts()
        if accounts and len(accounts) == 1:
            account = accounts[0]

            LOG.info(
                "Found account %s. Attempting to acquire token silently",
                account["username"],
            )
            token = self._app.acquire_token_silent(
                scopes=self.scope,
                account=account,
                force_refresh=True,
            )
        else:
            token = None

        if token is None:
            LOG.info("No suitable token found. Acquiring token interactively")
            res = self._app.acquire_token_interactive(
                scopes=self.scope,
            )
            if "error" in res:
                raise OdpAuthError("{}: {}".format(res["error"], res["error_description"]))

            # Token acquired successfully, rerun the method to get the token
            return self.authenticate()

        return token

    @staticmethod
    def _build_persistence(location: Path, fallback_to_plaintext: bool = False):
        if sys.platform.startswith("win"):
            return msal_extensions.FilePersistenceWithDataProtection(location)
        elif sys.platform.startswith("darwin"):
            return msal_extensions.KeychainPersistence(location, "odcat", "odcat-user")
        elif sys.platform.startswith("linux"):
            try:
                return msal_extensions.LibsecretPersistence(
                    str(location),
                    schema_name="odcat-sdk",
                    attributes={"app": "odcat", "component": "odcat-cli"},
                )
            except Exception:
                if not fallback_to_plaintext:
                    raise
                LOG.warning("Encryption unavailable. Opting in to plaintext")
        return msal_extensions.FilePersistence(str(location))



import typing

from odp.util.cheapdantic import BaseModel


class ColumnStats(BaseModel):
    """ColumnStats contains statistics about a column in a table"""

    name: str
    """name of the column"""

    metadata: typing.Optional[dict]
    """metadata of the column, if any"""

    type: str
    """type of the column, e.g., 'string', 'int', 'float', etc."""

    null_count: int
    """number of null values in the column"""

    num_values: int
    """number of non-null values in the column"""

    min: typing.Union[int, float, str, None]
    """minimum value in the column, if applicable"""

    max: typing.Union[int, float, str, None]
    """maximum value in the column, if applicable"""


class TableStats(BaseModel):
    """
    TableStats contains statistics about a table, including metadata, number of rows, columns.
    """

    num_rows: int
    """number of rows in the table"""

    size: int
    """size of the table in bytes, including metadata and schema"""

    columns: typing.Optional[list[ColumnStats]]



import io
import json
import logging
import os
from typing import TYPE_CHECKING, Any, Dict, Iterator, Optional, Union

import requests
from requests.adapters import HTTPAdapter
from urllib3 import Retry

from .util.check_version import get_sdk_version

# NOTE(oha): deferred imports to avoid circular dependencies, will be fixed when we remove the old client
# from odp.client.auth import InteractiveTokenProvider

if TYPE_CHECKING:
    from .dataset import Dataset

read_retry_strategy = Retry(
    total=8,
    status_forcelist=[429, 502, 503, 504],
    allowed_methods=["HEAD", "GET", "OPTIONS", "POST"],
    backoff_factor=1,  # exponential sleep (1s, 2s, 4s…)
)

write_retry_strategy = Retry(
    total=8,
    status_forcelist=[429],
    allowed_methods=["HEAD", "GET", "OPTIONS", "POST"],
    backoff_factor=2,
)


class Response:
    def __init__(self, res: requests.Response):
        if res.status_code == 204:
            raise FileNotFoundError(res.text)
        if res.status_code == 400:
            raise ValueError(res.text)
        res.raise_for_status()
        self.res = res

    def reader(self):
        if self.res.raw is None:
            return io.BytesIO(self.res.content)
        return self.res.raw

    def iter(self) -> Iterator[bytes]:
        if self.res.raw is None:
            return iter([self.res.content])
        return self.res.raw

    def all(self) -> bytes:
        if self.res.raw is None:
            return self.res.content
        return self.res.raw.read()

    def json(self) -> Union[dict, list, None]:
        return self.res.json()

    @property
    def content(self) -> Union[bytes, Any]:
        return self.res.content

    @property
    def status_code(self) -> int:
        return self.res.status_code


class Client:
    """
    New ODP client for accessing datasets and other resources.
    supersedes the old odp.OdpClient
    """

    def __init__(self, base_url: str = "https://api.hubocean.earth/", jwt_bearer: str = "", api_key: str = ""):
        if api_key:
            self._auth = lambda: f"ApiKey {api_key}"
        elif jwt_bearer:
            if not jwt_bearer.startswith("Bearer "):
                jwt_bearer = "Bearer " + jwt_bearer
            self._auth = lambda: jwt_bearer

        elif os.getenv("ODP_API_KEY"):
            api_key = os.getenv("ODP_API_KEY")
            self._auth = lambda: f"ApiKey {api_key}"

        elif os.getenv("JUPYTERHUB_API_TOKEN"):

            def get_token():
                res = requests.post("http://localhost:8000/access_token")
                res.raise_for_status()
                token: str = res.json()["token"]
                return "Bearer " + token

            self._auth = get_token
        else:
            self._setup_jwt()

        self.base_url = base_url.rstrip("/")

        # FIXME: this is wrong, sessions can't be this long
        self._http_client = requests.Session()
        self._http_client.headers.setdefault("User-Agent", "odp-sdk-python")

    def _setup_jwt(self):
        """
        fire up the JWT setup process on the browser
        """
        client_id = os.getenv("ODP_CLIENT_ID", "f96fc4a5-195b-43cc-adb2-10506a82bb82")
        from odp.client.auth import InteractiveTokenProvider

        prov = InteractiveTokenProvider(client_id=client_id)
        self._auth = lambda: f"Bearer {prov.get_token()}"
        # raise NotImplementedError("JWT authentication is not implemented yet.")

    def request(
        self,
        path: str,
        method: str = "POST",
        params: Optional[dict] = None,
        data: Union[Dict, bytes, Iterator[bytes], io.IOBase, None] = None,
        retry: bool = False,
    ) -> Response:
        """
        send a request to ODP
        """
        req = requests.Request(method, self.base_url + path, params=params, data=data)
        res = self._request(req, retry=retry)

        return Response(res)

    def _request(self, req: requests.Request, retry: bool = True) -> requests.Response:
        """
        base implementation to send a request to ODP
        may be overridden to use mechanisms other than http
        """
        req.headers.setdefault("Authorization", self._auth())  # allow override
        req.headers.setdefault("User-Agent", "odp-sdk-python/" + get_sdk_version())
        if isinstance(req.data, dict):
            req.data = json.dumps(req.data)
            req.headers.setdefault("Content-Type", "application/json")
        preq = req.prepare()
        # logging.info("request: %s %s %s", preq.method, preq.url, preq.body)
        if retry:
            retry_strategy = read_retry_strategy
        else:
            retry_strategy = write_retry_strategy

        with requests.Session() as s:
            adapter = HTTPAdapter(pool_connections=4, pool_maxsize=4, max_retries=retry_strategy)
            s.mount("http://", adapter)
            s.mount("https://", adapter)
            res = s.send(preq, stream=True)
        # this is the time it takes for the response headers to be returned, the body might take longer...
        logging.debug("response: %s in %.2fs from %s", res.status_code, res.elapsed.total_seconds(), res.url)
        return res

    def dataset(self, id: str) -> "Dataset":
        """fetch a dataset by id"""
        from .dataset import Dataset

        return Dataset(self, id)



from typing import Callable

import pyarrow as pa
import shapely


def has_geometry(schema: pa.Schema) -> str | list | None:
    """
    Returns the first field that is geometry, or a list of [lonfield, latfield] if both are found
    or None if no geometry fields are found
    """
    latfield = None
    lonfield = None
    for field in schema:
        meta = field.metadata or {}
        if b"isGeometry" in meta:
            return field.name
        if b"class" in meta:
            if meta[b"class"] == b"geometry":
                return field.name
            if meta[b"class"] == b"latitude":
                latfield = field.name
            elif meta[b"class"] == b"longitude":
                lonfield = field.name
    if latfield is not None and lonfield is not None:
        return [lonfield, latfield]
    return None


def geo_decoder(schema: pa.Schema) -> Callable[[dict], shapely.Geometry | None]:
    """
    Returns a function that decodes a row dict into a shapely geometry
    """
    lat_field, lon_field = None, None
    for field in schema:
        meta = field.metadata or {}
        if b"isGeometry" in meta:
            if pa.types.is_binary(field.type):

                def f(row) -> shapely.Geometry | None:
                    g = row[field.name]
                    del row[field.name]
                    if g is None:
                        return None
                    return shapely.from_wkb(g)

                return f
            if pa.types.is_string(field.type):

                def f(row) -> shapely.Geometry | None:
                    g = row[field.name]
                    del row[field.name]
                    if g is None:
                        return None
                    return shapely.from_wkt(g)

                return f
            raise ValueError(f"unsupported geometry type {field.type}")
        if b"class" in meta:
            if meta[b"class"] == b"latitude":
                lat_field = field.name
            elif meta[b"class"] == b"longitude":
                lon_field = field.name
    if lat_field is not None and lon_field is not None:

        def f(row) -> shapely.Geometry | None:
            lat, lon = row[lat_field], row[lon_field]
            # remove lat/lon fields from the properties
            del row[lat_field]
            del row[lon_field]
            if lat is None or lon is None:
                return None
            return shapely.Point(lon, lat)

        return f
    raise ValueError("no geometry field found")


def to_geodataframe(table: pa.Table):
    """
    Convert a pyarrow table to a GeoPandas geodataframe
    """
    try:
        import geopandas as gpd
    except Exception as e:
        raise ImportError("geopandas is required for to_geodataframe()") from e

    geom_spec = has_geometry(table.schema)
    if geom_spec is None:
        # geodf without active geometry
        return gpd.GeoDataFrame(table.to_pandas())

    decoder = geo_decoder(table.schema)
    rows = table.to_pylist()
    geoms = [decoder(r) for r in rows]

    df = table.to_pandas()

    if isinstance(geom_spec, str):
        drop_cols = [geom_spec] if geom_spec in df.columns else []
    elif isinstance(geom_spec, (list, tuple)) and len(geom_spec) == 2:
        drop_cols = [c for c in geom_spec if c in df.columns]
    else:
        drop_cols = []

    if drop_cols:
        df = df.drop(columns=drop_cols)

    return gpd.GeoDataFrame(df, geometry=geoms, crs="EPSG:4326")



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

import requests

from odp.new_client import Client


@dataclass(frozen=True)
class DatasetMeta:
    id: str
    name: str
    description: str

    @classmethod
    def from_dict(cls, data: dict) -> DatasetMeta:
        return DatasetMeta(
            id=data["id"],
            name=data["name"],
            description=data.get("description", ""),
        )


def create_dataset_meta(cli: Client, name: str, description: str = "n/a") -> DatasetMeta:
    """
    Create a new dataset with the given name and description.
    Returns the created Dataset object.
    """
    res = cli._request(
        requests.Request(
            method="POST",
            url=cli.base_url + "/api/catalog/v2/datasets",
            data={
                "name": name,
                "description": description,
            },
        ),
        retry=False,
    )
    res.raise_for_status()
    data = res.json()
    return DatasetMeta.from_dict(data)


def get_dataset_meta_by_name(cli: Client, name: str) -> Optional[DatasetMeta]:
    """
    Fetch a dataset by its name.
    Returns None if the dataset does not exist.
    """
    res = cli._request(
        requests.Request(
            method="GET",
            url=cli.base_url + "/api/catalog/v2/datasets",
        )
    )
    res.raise_for_status()
    for item in res.json():
        if item["name"] == name:
            return DatasetMeta.from_dict(item)

    return None


def get_dataset_meta_by_uuid(cli: Client, uuid: str) -> Optional[DatasetMeta]:
    """
    Fetch a dataset by its UUID.
    Returns None if the dataset does not exist.
    """
    res = cli._request(
        requests.Request(
            method="GET",
            url=cli.base_url + f"/api/catalog/v2/datasets/{uuid}",
        )
    )
    if res.status_code == 404:
        return None

    data = res.json()
    return DatasetMeta.from_dict(data)
